/* main.cpp
 * sets up the GUI and connects the main callback functions.
 *
 * for Denemo, a gtk+ frontend to GNU Lilypond
 * (c) 1999, 2000, 2001, 2002 Matthew Hiller, Adam Tee
 */
#include "config.h"
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <gtk/gtk.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_WAIT_H
#include <wait.h>
#endif
#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

/*snooper function for testing purposes */
static gint snooper (GtkWidget * grab_widget, GdkEventKey * event,
		     gpointer func_data);

GList *displays = NULL;
#include "view.h"
#include "exportxml.h"
#include "runsilent.h"
/* signal handler to be invoked when child processes _exit() without
 * having to wait for them */

/* Code by Erik Mouw, taken directly from the gtk+ FAQ */

void
sigchld_handler (gint num)
{
  sigset_t set, oldset;
  pid_t pid;
  gint status, exitstatus;

#ifndef G_OS_WIN32
  /* block other incoming SIGCHLD signals */
  sigemptyset (&set);
  sigaddset (&set, SIGCHLD);
  sigprocmask (SIG_BLOCK, &set, &oldset);

  /* wait for child */
  while ((pid = waitpid ((pid_t) - 1, &status, WNOHANG)) > 0)
    {
      if (WIFEXITED (status))
	{
	  exitstatus = WEXITSTATUS (status);

	  fprintf (stderr,
		   _("Parent: child exited, pid = %d, exit status = %d\n"),
		   (int) pid, exitstatus);
	}
      else if (WIFSIGNALED (status))
	{
	  exitstatus = WTERMSIG (status);

	  fprintf (stderr,
		   _("Parent: child terminated by signal %d, pid = %d\n"),
		   exitstatus, (int) pid);
	}
      else if (WIFSTOPPED (status))
	{
	  exitstatus = WSTOPSIG (status);

	  fprintf (stderr,
		   _("Parent: child stopped by signal %d, pid = %d\n"),
		   exitstatus, (int) pid);
	}
      else
	{
	  fprintf (stderr,
		   _("Parent: child exited magically, pid = %d\n"),
		   (int) pid);
	}
    }

  /* re-install the signal handler (some systems need this) */
  signal (SIGCHLD, sigchld_handler);

  /* and unblock it */
  sigemptyset (&set);
  sigaddset (&set, SIGCHLD);
  sigprocmask (SIG_UNBLOCK, &set, &oldset);
#endif
}
#if GTK_MAJOR_VERSION > 1
void segdialog(gchar *sigtype, gchar *message)
{
 GtkWidget *dialog;
 dialog = gtk_message_dialog_new (NULL, GTK_DIALOG_DESTROY_WITH_PARENT,
     				  GTK_MESSAGE_ERROR,GTK_BUTTONS_CLOSE,
				  "%s : %s", sigtype, message);
 gtk_dialog_run (GTK_DIALOG (dialog));
 gtk_widget_destroy (dialog);
}
#endif

//SIGSEGV Handler to do nice things if denemo bombs
void denemo_signal_handler(int sig)
{
  GList *tmp=NULL;
  struct scoreinfo *si;
  g_print("\nNo of displays : %d\n", g_list_length(displays)); 
  if(g_list_length(displays) == 1) {
    si = (struct scoreinfo *) displays->data;
    g_print("si is %p", si);
    char *filename = "bombedout.denemo";
    
    exportXML(filename, si, 0, 0);
  }else {
    int i=0;
     for(tmp=displays; tmp && g_list_length(tmp) > 1; tmp=tmp->next) {
       si = (struct scoreinfo *)tmp->data;
       char *filename = "bombedout";
       char t[5];
       sprintf(t, "%d", i); 
       strcat(filename,t); 
       strcat(filename,".denemo");
       exportXML(filename, si, 0, 0);
       i++;
     }
  }
     
  switch(sig) {
    case SIGSEGV:
#if GTK_MAJOR_VERSION > 1
      segdialog("Seg fault","Unable to continue"); 
#endif
      g_print("Seg fault : Unable to continue\n");
      break;
  }
  exit(1);
}
      


int
main (int argc, char *argv[])
{

  gint opts;
#ifdef HAVE_GETOPT_H
  static struct option long_options[] = {
    {"help", no_argument, NULL, 'h'},
    {"test", required_argument, NULL, 't'},
    {"version", no_argument, NULL, 'v'}
  };
#endif
  gchar *helptext = g_strconcat (_("\nGNU Denemo version "), VERSION ".\n\n",
				 _("usage: denemo [OPTION... [FILE]\n\n\
Run denemo, opening save file FILE\n\n\
Denemo is a graphical music notation editor.  It produces save files\n\
in GNU Lilypond input format (suitable for immediate typesetting with GNU\n\
Lilypond) and Adam Tee's JTF file format. Denemo is part of the GNU\n\
project.\n\n\
Options:\n\
  -h,--help                 print this help and exit\n\
  -t,--test <testfile>      execute/create test\n\
  -s,--silent		    lets just start with silent lilypond conversion\n\
  -v,--version              print version number and exit\n\n\n\
Report bugs to bug-denemo@gnu.org\n"), NULL);

  gchar *copytext
    = _("(c) 1999, 2000, 2001 Matthew Hiller, Adam Tee, and others\n\n\n\
This program is provided with absolutely NO WARRANTY; see\n\
the file COPYING for details.\n\n\
This software may be redistributed and modified under the\n\
terms of the GNU General Public License; again, see the file\n\
COPYING for details.\n\n");
  /* are we in test mode (recording or playback) */
  gboolean denemo_test_mode = FALSE;
  /* name of testfile, containing keypresses to be played back for automated testing */
  gchar *testfile = NULL;
  /*  gchar *win32text
     = _("Warning: the win32 port of Denemo is as yet incomplete.\n\
     Most prominently, playback and options do not work.\n\n"); */

  gtk_init (&argc, &argv);

  //segdialog("Test", "Hello");
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);

#ifndef G_OS_WIN32
#ifdef HAVE_GETOPT_H
  while ((opts = getopt_long (argc, argv, "shvt:", long_options, NULL)) != -1)
#else
    while ((opts = getopt(argc, argv, "shvt:")) != -1)
#endif
    {
      if (opts == 'h')
	{
	  printf (helptext);
	  exit (0);
	}
      else if (opts == 't')
	{
	  denemo_test_mode = TRUE;
	  testfile = g_strdup ((const gchar *) optarg);
	  g_warning
	    ("Test rig: keypresses will be read from/written to file %s\n",
	     testfile);
	}
      else if (opts == 's')
        {
	   struct scoreinfo *si = 
	     (struct scoreinfo *)g_malloc(sizeof(struct scoreinfo));
	   printf(copytext);
	   silentconversion(argv[optind], si);
	   exit(0);
	}
      else if (opts == 'v')
	{
	  printf (_("\nGNU Denemo version "));
	  printf (VERSION ".\n\n");
	  printf (copytext);
	  exit (0);
	}
    }
#endif

  printf (_("\nGNU Denemo, a gtk+ frontend for GNU Lilypond\n"));
  printf (copytext);

#ifdef G_OS_WIN32
  printf (win32text);
#endif
  g_free (helptext);

  midi_init();

  newview();
/* Set up the signal handler */
#ifndef G_OS_WIN32
    signal (SIGSEGV, denemo_signal_handler);
    signal (SIGCHLD, sigchld_handler);
#endif
      
  /* And open a file, if it was specified on the command line. Note
   * that this had to be done after the window was created, otherwise
   * there wouldn't have been a titlebar to set. Also note that
   * a blank score is created whether or not a load was specified.
   * This is done this way because the load could bomb out. */

#ifndef G_OS_WIN32
  if (optind < argc)
    {
      GList *tmp = g_list_nth (displays, 0);
      struct scoreinfo *si = (struct scoreinfo *) tmp->data;
      open_for_real (argv[optind], si);
    }
#endif


  if (denemo_test_mode)
    gtk_key_snooper_install (snooper, (gpointer) testfile);

  /* Now launch into the main gtk event loop and we're all set */
  gtk_main ();
  return 0;

}

/* snooper function installed by gtk_key_snooper_install (snooper, testfilename)
 * if a file exists of name testfilename then it supplies keypresses to substitute
 * for the ones typed by the user - ie it "plays back" the test. If not it creates
 * a new test by storing each keypress from the user into a file of that name.
 * Note if keypresses are sent too fast they may go
 * to the wrong window, and the test will appear to fail.
*/
static gint
snooper (GtkWidget * grab_widget, GdkEventKey * event, gpointer func_data)
{
  static FILE *fp;		/* file containing keypress events */
  static gboolean recording = FALSE;	/* we are recording keystrokes */
  gboolean ignore_this_key = FALSE;
  GdkWindow *window;
  guint32 time;
  gchar *string;
  gchar *testfile = (gchar *) func_data;	/* name of test file containing keypresses */

  window = event->window;
  time = event->time;
  if (!fp)
    fp = fopen (testfile, "r");
  if (!fp)			/* it is a new file, so we are recording a new test */
    {
      fp = fopen (testfile, "w");
      recording = TRUE;
    }
  g_assert (fp);		/* can't open the file */
  if (recording)
    {
      fwrite (event, sizeof (*event), 1, fp);
      fwrite (event->string, event->length, 1, fp);
    }
  else				/* playing back keystrokes from file */
    {
      size_t check;		/* check the read from file for this event */
      check = fread (event, sizeof (*event), 1, fp);
      g_assert (check == 1);
      string = (gchar *) g_malloc (event->length + 1);	/* memory leak? - ok for test rig */
      g_assert (string);
      if (event->length)
	check = fread (string, event->length, 1, fp);
      g_assert (check == 1);
      *(string + event->length) = 0;	/* null terminate */

      /* freshen the stale pointers and time */
      event->window = window;
      event->time = time;
      event->string = string;
    }

  return ignore_this_key;	/* FALSE => allow the keypress to be acted on normally */
}
